from collections import deque
import queue
import threading

from . import _rawapi
from .. import util


class WIAException(util.PyinsaneException):
    def __init__(self, msg):
        super(WIAException, self).__init__("WIA: {}".format(msg))


class WiaAction(object):
    def __init__(self, func, **kwargs):
        self.func = func
        self.kwargs = kwargs
        self.result = None
        self.exception = None
        self.__sem = threading.Semaphore(0)

    def start(self):
        global wia_thread
        global wia_action_queue

        if wia_thread is None or not wia_thread.is_alive():
            raise WIAException("WIA thread died unexpectidly !")
        wia_action_queue.put(self)

    def wait(self):
        self.start()
        self.__sem.acquire()
        if self.exception is not None:
            raise self.exception
        return self.result

    def do(self):
        try:
            self.result = self.func(**self.kwargs)
        except Exception as exc:
            self.exception = exc
        self.__sem.release()


class WiaWorker(threading.Thread):
    def run(self):
        global wia_action_queue
        while True:
            try:
                action = wia_action_queue.get(block=True, timeout=1)
                action.do()
            except queue.Empty:
                if not parent_thread.is_alive():
                    return


parent_thread = threading.current_thread()
wia_action_queue = queue.Queue()
wia_thread = WiaWorker()
wia_thread.start()


def _init():
    _rawapi.init()


def init():
    return WiaAction(_init).wait()


def _open(devid):
    out = _rawapi.open(devid)
    if out is None:
        raise WIAException("Failed to open {}".format(devid))
    return out


def open(devid):
    return WiaAction(_open, devid=devid).wait()


def _get_devices():
    devices = _rawapi.get_devices()
    if devices is None:
        raise WIAException("Failed to get device list")
    return devices


def get_devices():
    return WiaAction(_get_devices).wait()


def _get_sources(dev):
    sources = _rawapi.get_sources(dev)
    if sources is None:
        raise WIAException("Failed to get sources")
    return sources


def get_sources(dev):
    return WiaAction(_get_sources, dev=dev).wait()


def _get_properties(dev_or_src):
    properties = _rawapi.get_properties(dev_or_src)
    if properties is None:
        raise WIAException("Failed to get scanner properties")
    return properties


def get_properties(dev_or_src):
    return WiaAction(_get_properties, dev_or_src=dev_or_src).wait()


def _get_constraints(dev_or_src):
    constraints = _rawapi.get_constraints(dev_or_src)
    if constraints is None:
        raise WIAException("Failed to get properties constraints")
    return constraints


def get_constraints(dev_or_src):
    return WiaAction(_get_constraints, dev_or_src=dev_or_src).wait()


def _set_property(dev_or_src, propname, propvalue):
    ret = _rawapi.set_property(dev_or_src, propname, propvalue)
    if not ret:
        raise WIAException("Failed to set scanner properties")


def set_property(dev_or_src, propname, propvalue):
    return WiaAction(_set_property, dev_or_src=dev_or_src,
                     propname=propname, propvalue=propvalue).wait()


class WiaCallbacks(object):
    def __init__(self):
        super(WiaCallbacks, self).__init__()
        self.received = deque()
        self.condition = threading.Condition()
        self.buffer = 512000 * b"\0"

    def get_data_cb(self, nb_bytes):
        self.condition.acquire()
        try:
            data = self.buffer[:nb_bytes]
            self.received.append(data)
            self.condition.notify_all()
        finally:
            self.condition.release()

    def end_of_page_cb(self):
        self.condition.acquire()
        try:
            self.received.append(
                EOFError()
            )
            self.condition.notify_all()
        finally:
            self.condition.release()

    def end_of_scan_cb(self):
        self.condition.acquire()
        try:
            self.received.append(
                StopIteration()
            )
            self.condition.notify_all()
        finally:
            self.condition.release()

    def read(self):
        self.condition.acquire()
        try:
            if len(self.received) <= 0:
                self.condition.wait()
            popped = self.received.popleft()
            if isinstance(popped, Exception):
                raise popped
            return popped
        finally:
            self.condition.release()


def _start_scan(src, out):
    ret = _rawapi.download(
        src,
        out.get_data_cb,
        out.end_of_page_cb,
        out.end_of_scan_cb,
        out.buffer,
    )
    if ret is None:
        raise WIAException("Failed to start scan")
    return ret


def start_scan(src):
    out = WiaCallbacks()
    WiaAction(_start_scan, src=src, out=out).start()  # don't wait
    return out


def exit():
    _rawapi.exit()



import logging
import os
import pickle
import shutil
import struct
import sys
import tempfile

import PIL.Image

# import basic elements directly, so the caller
# doesn't have to import rawapi if they need them.
from . import abstract
from .rawapi import SaneCapabilities
from .rawapi import SaneConstraint
from .rawapi import SaneConstraintType
from .rawapi import SaneException
from .rawapi import SaneStatus
from .rawapi import SaneUnit
from .rawapi import SaneValueType


__all__ = [
    'SaneCapabilities',
    'SaneConstraint',
    'SaneConstraintType',
    'SaneException',
    'SaneStatus',
    'SaneValueType',
    'SaneUnit',
    'init',
    'exit',
    'Scanner',
    'ScannerOption',
    'get_devices',
]


logger = logging.getLogger(__name__)

length_size = None
fifo_s2c = None
fifo_c2s = None
pipe_path_c2s = None
pipe_path_s2c = None
pipe_dirpath = None


def remote_do(command, *args, **kwargs):
    global length_size
    global fifo_s2c
    global fifo_c2s
    global pipe_path_c2s
    global pipe_path_s2c

    cmd = {
        'command': command,
        'args': args,
        'kwargs': kwargs,
    }

    cmd = pickle.dumps(cmd)
    length = struct.pack("i", len(cmd))
    os.write(fifo_c2s, length)
    os.write(fifo_c2s, cmd)

    length = os.read(fifo_s2c, length_size)
    length = struct.unpack("i", length)[0]
    result = os.read(fifo_s2c, length)
    assert(len(result) == length)
    result = pickle.loads(result)
    if 'exception' in result:
        exc_class = eval(result['exception'])
        raise exc_class(*result['exception_args'])

    if command == 'exit':
        os.close(fifo_c2s)
        os.close(fifo_s2c)
        os.unlink(pipe_path_c2s)
        os.unlink(pipe_path_s2c)
        shutil.rmtree(pipe_dirpath)
        return

    return result['out']


def init():
    global length_size
    global fifo_s2c
    global fifo_c2s
    global pipe_path_c2s
    global pipe_path_s2c
    global pipe_dirpath

    start_daemon = os.getenv('PYINSANE_DAEMON', '1')
    start_daemon = True if int(start_daemon) > 0 else False

    if not start_daemon:
        return

    logger.info("Starting Pyinsane subprocess")

    pipe_dirpath = tempfile.mkdtemp(prefix="pyinsane_")
    pipe_path_c2s = os.path.join(pipe_dirpath, "pipe_c2s")
    os.mkfifo(pipe_path_c2s)
    pipe_path_s2c = os.path.join(pipe_dirpath, "pipe_s2c")
    os.mkfifo(pipe_path_s2c)

    logger.info("Pyinsane pipes: {} | {}".format(pipe_path_c2s, pipe_path_s2c))

    if os.fork() == 0:
        # prevent the daemon from starting itself (due to the way
        # imports behave)
        os.putenv('PYINSANE_DAEMON', '0')
        os.execlp(
            sys.executable, sys.executable,
            "-m", "pyinsane2.sane.daemon",
            pipe_dirpath,
            pipe_path_c2s, pipe_path_s2c
        )

    length_size = len(struct.pack("i", 0))
    fifo_c2s = os.open(pipe_path_c2s, os.O_WRONLY)
    fifo_s2c = os.open(pipe_path_s2c, os.O_RDONLY)

    logger.info("Connected to Pyinsane subprocess")


def exit():
    remote_do('exit')


class ScannerOption(object):
    _abstract_opt = None
    _scanner_name = None

    idx = 0
    name = ""
    title = ""
    desc = ""
    val_type = SaneValueType(SaneValueType.INT)
    unit = SaneUnit(SaneUnit.NONE)
    size = 4
    capabilities = SaneCapabilities(SaneCapabilities.NONE)

    constraint_type = SaneConstraintType(SaneConstraintType.NONE)
    constraint = None

    def __init__(self, scanner, idx):
        self.idx = idx
        self._abstract_opt = abstract.ScannerOption(scanner._abstract_dev, idx)

    @staticmethod
    def build_from_abstract(scanner, abstract_opt):
        opt = ScannerOption(scanner, abstract_opt.idx)
        opt._scanner_name = scanner.name
        opt._abstract_opt = abstract_opt
        opt.name = abstract_opt.name
        opt.title = abstract_opt.title
        opt.desc = abstract_opt.desc
        opt.val_type = abstract_opt.val_type
        opt.unit = abstract_opt.unit
        opt.size = abstract_opt.size
        opt.capabilities = abstract_opt.capabilities
        opt.constraint_type = abstract_opt.constraint_type
        opt.constraint = abstract_opt.constraint
        return opt

    def _get_value(self):
        return remote_do('get_option_value', self._scanner_name, self.name)

    def _set_value(self, new_value):
        remote_do('set_option_value', self._scanner_name, self.name, new_value)

    value = property(_get_value, _set_value)


class Scan(object):
    def __init__(self, scanner_name):
        self._scanner_name = scanner_name

    def read(self):
        return remote_do('scan_read', self._scanner_name)

    def _get_available_lines(self):
        return remote_do('scan_get_available_lines', self._scanner_name)

    available_lines = property(_get_available_lines)

    def _get_expected_size(self):
        return remote_do('scan_get_expected_size', self._scanner_name)

    expected_size = property(_get_expected_size)

    def get_image(self, start_line=0, end_line=-1):
        img = remote_do(
            'scan_get_image', self._scanner_name, start_line, end_line
        )
        return PIL.Image.frombytes(*img)

    def cancel(self):
        return remote_do('scan_cancel', self._scanner_name)


class ScanSession(object):
    def __init__(self, scanner, multiple=False):
        self._scanner = scanner.name
        self._remote_session = remote_do('scan', scanner.name, multiple)
        self.scan = Scan(scanner.name)

    def __get_imgs(self):
        imgs = remote_do('get_images', self._scanner)
        imgs = [PIL.Image.frombytes(*img) for img in imgs]
        return imgs

    images = property(__get_imgs)

    def read(self):
        """
        Deprecated
        """
        self.scan.read()

    def get_nb_img(self):
        """
        Deprecated
        """
        return len(self.images)

    def get_img(self, idx=0):
        """
        Deprecated
        """
        return self.images[idx]


class Scanner(object):
    def __init__(self, name=None,
                 vendor="Unknown", model="Unknown", dev_type="Unknown",
                 abstract_dev=None):
        if abstract_dev is None:
            abstract_dev = abstract.Scanner(name)
        else:
            vendor = abstract_dev.vendor
            model = abstract_dev.model
            dev_type = abstract_dev.dev_type
        self._abstract_dev = abstract_dev
        self.name = name
        self.nice_name = name  # for WIA compatibility
        self.vendor = vendor
        self.model = model
        self.dev_type = dev_type

    @staticmethod
    def build_from_abstract(abstract_dev):
        return Scanner(abstract_dev.name, abstract_dev=abstract_dev)

    def _get_options(self):
        options = remote_do("get_options", self.name)
        return {
            x.name: ScannerOption.build_from_abstract(self, x)
            for x in options.values()
        }

    options = property(_get_options)

    def scan(self, multiple=False):
        return ScanSession(self, multiple)

    def __str__(self):
        return ("'%s' (%s, %s, %s)"
                % (self.name, self.vendor, self.model, self.dev_type))


def get_devices(local_only=False):
    return [
        Scanner.build_from_abstract(x)
        for x in remote_do('get_devices', local_only)
    ]



import os
import unittest

if os.name == "nt":
    from pyinsane2.wia import rawapi


class TestInit(unittest.TestCase):
    def setUp(self):
        pass

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_init(self):
        version = rawapi.init()
        rawapi.exit()

    def tearDown(self):
        pass


class TestGetDevices(unittest.TestCase):
    def setUp(self):
        if os.name == "nt":
            rawapi.init()

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_get_devices(self):
        devices = rawapi.get_devices()
        self.assertTrue(len(devices) > 0)

    def tearDown(self):
        if os.name == "nt":
            rawapi.exit()


class TestOpenDevice(unittest.TestCase):
    def setUp(self):
        if os.name == "nt":
            rawapi.init()

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_open_device(self):
        devices = rawapi.get_devices()
        self.assertTrue(len(devices) > 0)
        devid = devices[0][0]
        dev = rawapi.open(devid)
        self.assertNotEqual(dev, None)
        del dev
        dev = None

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_invalid_open_device(self):
        self.assertRaises(rawapi.WIAException, rawapi.open, "randomcraphere")

    def tearDown(self):
        if os.name == "nt":
            rawapi.exit()


class TestGetSources(unittest.TestCase):
    def setUp(self):
        if os.name != "nt":
            return
        rawapi.init()

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_get_sources(self):
        devices = rawapi.get_devices()
        self.assertTrue(len(devices) > 0)
        devid = devices[0][0]
        dev = rawapi.open(devid)
        self.assertNotEqual(dev, None)
        sources = rawapi.get_sources(dev)
        self.assertTrue(len(sources) > 0)

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_invalid_get_sources(self):
        self.assertRaises(
            rawapi.WIAException, rawapi.get_sources, "randomcrappyobject"
        )

    def tearDown(self):
        if os.name != "nt":
            return
        rawapi.exit()


class TestGetProperties(unittest.TestCase):
    def setUp(self):
        if os.name != "nt":
            return
        rawapi.init()
        devices = rawapi.get_devices()
        devid = devices[0][0]
        self.dev = rawapi.open(devid)
        self.sources = rawapi.get_sources(self.dev)
        self.assertTrue(len(self.sources) > 0)

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_get_dev_properties(self):
        props = rawapi.get_properties(self.dev)
        self.assertTrue(len(props) > 0)

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_get_src_properties(self):
        props = rawapi.get_properties(self.sources[0][1])
        self.assertTrue(len(props) > 0)

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_invalid_get_properties(self):
        self.assertRaises(rawapi.WIAException, rawapi.get_properties, "crappy_obj")

    def tearDown(self):
        if os.name != "nt":
            return
        rawapi.exit()


class TestSetProperty(unittest.TestCase):
    def setUp(self):
        if os.name != "nt":
            return
        rawapi.init()
        devices = rawapi.get_devices()
        devid = devices[0][0]
        self.dev = rawapi.open(devid)
        self.sources = rawapi.get_sources(self.dev)
        self.assertTrue(len(self.sources) > 0)

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_set_src_property_depth(self):
        rawapi.set_property(self.sources[0][1], "depth", 8)

        props = rawapi.get_properties(self.sources[0][1])
        self.assertTrue(len(props) > 0)
        for (propname, propvalue, accessright, _) in props:
            self.assertTrue(accessright == "ro" or accessright == "rw")
            if (propname == "depth"):
                self.assertEqual(propvalue, 8)
                self.assertEqual(accessright, "rw")

        rawapi.set_property(self.sources[0][1], "depth", 24)

        props = rawapi.get_properties(self.sources[0][1])
        self.assertTrue(len(props) > 0)
        for (propname, propvalue, _a, _b) in props:
            if (propname == "depth"):
                self.assertEqual(propvalue, 24)

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_set_src_property_preview(self):
        rawapi.set_property(self.sources[0][1], "preview", "preview_scan")

        props = rawapi.get_properties(self.sources[0][1])
        self.assertTrue(len(props) > 0)
        for (propname, propvalue, accessright, _) in props:
            self.assertTrue(accessright == "ro" or accessright == "rw")
            if (propname == "preview"):
                self.assertEqual(propvalue, "preview_scan")
                self.assertEqual(accessright, "rw")

        rawapi.set_property(self.sources[0][1], "preview", "final_scan")

        props = rawapi.get_properties(self.sources[0][1])
        self.assertTrue(len(props) > 0)
        for (propname, propvalue, _a, _b) in props:
            if (propname == "preview"):
                self.assertEqual(propvalue, "final_scan")

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_invalid_set_src_property_depth(self):
        self.assertRaises(rawapi.WIAException, rawapi.set_property, "crapobj", "depth", 8)
        self.assertRaises(rawapi.WIAException, rawapi.set_property, self.sources[0][1], "crapproperty", 8)
        self.assertRaises(rawapi.WIAException, rawapi.set_property, self.sources[0][1], "depth", "crapvalue")

    def tearDown(self):
        if os.name != "nt":
            return
        rawapi.exit()


class TestScan(unittest.TestCase):
    def setUp(self):
        if os.name != "nt":
            return
        rawapi.init()
        devices = rawapi.get_devices()
        devid = devices[0][0]
        self.dev = rawapi.open(devid)
        self.sources = rawapi.get_sources(self.dev)
        self.assertTrue(len(self.sources) > 0)

    @unittest.skipIf(os.name != "nt", "Windows only")
    def test_scan(self):
        scan = rawapi.start_scan(self.sources[0][1])
        try:
            while True:
                buf = scan.read()
                self.assertTrue(len(buf) > 0)
        except EOFError:
            pass
        self.assertRaises(StopIteration, scan.read)  # no more than one page expected

    def tearDown(self):
        if os.name != "nt":
            return
        rawapi.exit()



import ctypes
import functools

from .. import util


__all__ = [
    'SaneCapabilities',
    'SaneConstraint',
    'SaneConstraintType',
    'SaneDevice',
    'SaneException',
    'SaneFrame',
    'SaneInfo',
    'SaneOptionDescriptor',
    'SaneParameters',
    'SaneRange',
    'SaneStatus',
    'SaneUnit',
    'SaneValueType',
    'SaneVersion',

    'is_sane_available',
    'sane_init',
    'sane_exit',
    'sane_get_devices',
    'sane_open',
    'sane_close',
    'sane_get_option_descriptor',
    'sane_get_option_value',
    'sane_set_option_value',
    'sane_set_option_auto',
    'sane_get_parameters',
    'sane_start',
    'sane_read',
    'sane_cancel',
    'sane_set_io_mode',
    'sane_get_select_fd',
    'sane_strstatus',
]


@functools.total_ordering
class SaneEnum(object):
    VALUE_TO_STR = {}

    def __init__(self, value):
        self.__value = value

    def __int__(self):
        return self.__value

    def __eq__(self, other):
        if isinstance(other, int):
            return self.__value == other
        return self.__value == other.__value

    def __lt__(self, other):
        if isinstance(other, int):
            return self.__value < other
        return self.__value < other.__value

    def __str__(self):
        if self.__value not in self.VALUE_TO_STR:
            txt = "Unknown value (%d)" % (self.__value)
        else:
            txt = "%s (%d)" % (self.VALUE_TO_STR[self.__value], self.__value)
        return "%s : %s" % (type(self), txt)


class SaneFlags(object):
    FLAG_TO_STR = {}

    def __init__(self, flags=0):
        self.__flags = flags

    def __int__(self):
        return self.__flags

    def __add__(self, new_flag):
        return (self.__flags | new_flag)

    def __sub__(self, old_flag):
        return (self.flags & ~(old_flag))

    def __contains__(self, flag):
        return ((self.__flags & flag) == flag)

    def __hex__(self):
        return hex(self.__flags)

    def __cmp__(self, other):
        return cmp(self.__flags, other.__flags)

    def __str__(self):
        txt = "%s :" % (type(self))
        txt += "["
        for flag in self.FLAG_TO_STR:
            if flag in self:
                txt += " %s," % (self.FLAG_TO_STR[flag])
        txt += "]"
        return txt


class SaneVersion(object):
    SANE_CURRENT_MAJOR = 1
    SANE_CURRENT_MINOR = 0

    def __init__(self, major, minor, build=0):
        self.major = major
        self.minor = minor
        self.build = build

    def is_current(self):
        return ((self.major == self.SANE_CURRENT_MAJOR) and
                (self.minor == self.SANE_CURRENT_MINOR))

    def __str__(self):
        return "Sane version: %d.%d.%d" % (self.major, self.minor, self.build)


class SaneStatus(SaneEnum):
    GOOD = 0
    UNSUPPORTED = 1
    CANCELLED = 2
    DEVICE_BUSY = 3
    INVAL = 4
    EOF = 5
    JAMMED = 6
    NO_DOCS = 7
    COVER_OPEN = 8
    IO_ERROR = 9
    NO_MEM = 10
    ACCESS_DENIED = 11
    WARMING_UP = 12
    HW_LOCKED = 13

    VALUE_TO_STR = {
        GOOD:          "No error",
        UNSUPPORTED:   "Operation is not supported",
        CANCELLED:     "Operation was cancelled",
        DEVICE_BUSY:   "Device is busy. Try again later",
        INVAL:         "Data is invalid",
        EOF:           "End-of-file : no more data available",
        JAMMED:        "Document feeder jammed",
        NO_DOCS:       "Document feeder out of documents",
        COVER_OPEN:    "Scanner cover is open",
        IO_ERROR:      "Error during device I/O",
        NO_MEM:        "Out of memory",
        ACCESS_DENIED: "Access to resource has been denied",
        WARMING_UP:    "Lamp is not ready yet. Try again later",
        HW_LOCKED:     "Scanner mechanism locked for transport",
    }


class SaneException(util.PyinsaneException):
    def __init__(self, status):
        Exception.__init__(self, str(status))
        self.status = status


class SaneValueType(SaneEnum):
    BOOL = 0
    INT = 1
    FIXED = 2
    STRING = 3
    BUTTON = 4
    GROUP = 5

    VALUE_TO_STR = {
        BOOL:      "Boolean",
        INT:       "Integer",
        FIXED:     "Fixed",
        STRING:    "String",
        BUTTON:    "Button",
        GROUP:     "Group",
    }

    VALUE_TO_CLASS = {
        BOOL:      ctypes.c_int,
        INT:       ctypes.c_int,
        FIXED:     ctypes.c_int,
        STRING:    ctypes.c_buffer,
    }

    def can_getset_opt(self):
        return int(self) in self.VALUE_TO_CLASS

    def buf_to_pyobj(self, buf):
        cl = self.VALUE_TO_CLASS[int(self)]
        if cl == ctypes.c_buffer:
            return buf.value
        return cl.from_buffer(buf).value

    def get_ctype_obj(self, pyobj):
        cl = self.VALUE_TO_CLASS[int(self)]
        return cl(pyobj)


class SaneUnit(SaneEnum):
    NONE = 0
    PIXEL = 1
    BIT = 2
    MM = 3
    DPI = 4
    PERCENT = 5
    MICROSECOND = 6

    VALUE_TO_STR = {
        NONE:          "None",
        PIXEL:         "Pixel",
        BIT:           "Bit",
        MM:            "Mm",
        DPI:           "Dpi",
        PERCENT:       "Percent",
        MICROSECOND:   "Microsecond",
    }


class SaneDevice(ctypes.Structure):
    _fields_ = [
        ("name", ctypes.c_char_p),
        ("vendor", ctypes.c_char_p),
        ("model", ctypes.c_char_p),
        ("type", ctypes.c_char_p),
    ]

    def __str__(self):
        return ("Device: %s (%s, %s, %d)"
                % (self.name, self.vendor, self.model, self.type))


class SaneCapabilities(SaneFlags):
    NONE = 0

    SOFT_SELECT = (1 << 0)
    HARD_SELECT = (1 << 1)
    SOFT_DETECT = (1 << 2)
    EMULATED = (1 << 3)
    AUTOMATIC = (1 << 4)
    INACTIVE = (1 << 5)
    ADVANCED = (1 << 6)

    FLAG_TO_STR = {
        SOFT_SELECT:   "Soft_select",
        HARD_SELECT:   "Hard_select",
        SOFT_DETECT:   "Soft_detect",
        EMULATED:      "Emulated",
        AUTOMATIC:     "Automatic",
        INACTIVE:      "Inactive",
        ADVANCED:      "Advanced",
    }

    def is_active(self):
        return self.INACTIVE not in self

    def is_settable(self):
        return self.SOFT_SELECT in self


class SaneInfo(SaneFlags):
    EMPTY = 0

    INEXACT = (1 << 0)
    RELOAD_OPTIONS = (1 << 1)
    RELOAD_PARAMS = (1 << 2)

    FLAG_TO_STR = {
        INEXACT:           "Inexact",
        RELOAD_OPTIONS:    "Reload_options",
        RELOAD_PARAMS:     "Reload_params",
    }


class SaneConstraintType(SaneEnum):
    NONE = 0
    RANGE = 1
    WORD_LIST = 2
    STRING_LIST = 3

    VALUE_TO_STR = {
        NONE:          "None",
        RANGE:         "Range",
        WORD_LIST:     "Word list",
        STRING_LIST:   "String list",
    }

    @staticmethod
    def __constraint_none_to_pyobj(sane_constraint):
        return None

    @staticmethod
    def __constraint_range_to_pyobj(sane_constraint):
        return (sane_constraint.range.contents.min,
                sane_constraint.range.contents.max,
                sane_constraint.range.contents.quant)

    @staticmethod
    def __constraint_word_list_to_pyobj(sane_constraint):
        list_lng = sane_constraint.word_list[0]
        return sane_constraint.word_list[1:list_lng+1]

    @staticmethod
    def __constraint_string_list_to_pyobj(sane_constraint):
        string_list = []
        idx = 0
        while sane_constraint.string_list[idx]:
            string = sane_constraint.string_list[idx]
            if hasattr(string, 'decode'):
                string = string.decode("utf-8")
            string_list.append(string)
            idx += 1
        return string_list

    def get_pyobj_constraint(self, sane_constraint):
        pyobj = {
            self.NONE:         self.__constraint_none_to_pyobj,
            self.RANGE:        self.__constraint_range_to_pyobj,
            self.WORD_LIST:    self.__constraint_word_list_to_pyobj,
            self.STRING_LIST:  self.__constraint_string_list_to_pyobj,
        }[int(self)](sane_constraint)
        return pyobj


class SaneRange(ctypes.Structure):
    _fields_ = [
        ("min", ctypes.c_int),
        ("max", ctypes.c_int),
        ("quant", ctypes.c_int),
    ]

    def __str__(self):
        return "Range: %d-%d (%d)" % (self.min, self.max, self.quant)


class SaneConstraint(ctypes.Union):
    _fields_ = [
        ("string_list", ctypes.POINTER(ctypes.c_char_p)),
        ("word_list", ctypes.POINTER(ctypes.c_int)),
        ("range", ctypes.POINTER(SaneRange))
    ]


class SaneOptionDescriptor(ctypes.Structure):
    _fields_ = [
        ("name", ctypes.c_char_p),
        ("title", ctypes.c_char_p),
        # Is actually multi-line ! -> you have to get
        # a pointer on the pointer and use ptr[0], ptr[1], etc to get all the
        # lines
        ("desc", ctypes.c_char_p),
        ("type", ctypes.c_int),  # SaneValueType
        ("unit", ctypes.c_int),  # SaneUnit
        ("size", ctypes.c_int),
        ("cap", ctypes.c_int),  # SaneCapabilities

        ("constraint_type", ctypes.c_int),  # SaneConstraintType
        ("constraint", SaneConstraint),
    ]


class SaneAction(SaneEnum):
    GET_VALUE = 0
    SET_VALUE = 1
    SET_AUTO = 3

    VALUE_TO_STR = {
        GET_VALUE: "Get value",
        SET_VALUE: "Set value",
        SET_AUTO:  "Set auto",
    }


class SaneFrame(SaneEnum):
    GRAY = 0
    RGB = 1
    RED = 2
    GREEN = 3
    BLUE = 4

    VALUE_TO_STR = {
        GRAY:  "Gray",
        RGB:   "RGB",
        RED:   "Red",
        GREEN: "Green",
        BLUE:  "Blue",
    }

    VALUE_TO_PIL_FORMAT = {
        GRAY: "L",
        RGB: "RGB",
        RED: "L",
        GREEN: "L",
        BLUE: "L",
    }

    def get_pil_format(self):
        return self.VALUE_TO_PIL_FORMAT[int(self)]


class SaneParameters(ctypes.Structure):
    _fields_ = [
        ("format",          ctypes.c_int),  # SaneFrame
        ("last_frame",      ctypes.c_int),  # boolean
        ("bytes_per_line",  ctypes.c_int),
        ("pixels_per_line", ctypes.c_int),
        ("lines",           ctypes.c_int),
        ("depth",           ctypes.c_int),
    ]


def __dummy_auth_callback(sane_ressource_str):
    return (
        "anonymous",  # login
        ""  # password
    )


class __AuthCallbackWrapper(object):
    MAX_USERNAME_LEN = 128
    MAX_PASSWORD_LEN = 128

    def __init__(self, auth_callback):
        self.__auth_callback = auth_callback

    def wrapper(self, ressource_ptr, login_ptr, passwd_ptr):
        (login, password) = self.__auth_callback(ressource_ptr.value)
        # TODO(Jflesch): Make sure the following works
        ctypes.memmove(login_ptr, ctypes.c_char_p(login),
                       min(len(login)+1, self.MAX_USERNAME_LEN))
        ctypes.memmove(passwd_ptr, ctypes.c_char_p(password),
                       min(len(password)+1, self.MAX_USERNAME_LEN))


sane_is_init = 0
sane_version = None

sane_available = False

for libname in ["libsane.so.1", "libsane.1.dylib"]:
    try:
        SANE_LIB = ctypes.cdll.LoadLibrary(libname)
        sane_available = True
        break
    except OSError:
        pass

if sane_available:
    AUTH_CALLBACK_DEF = ctypes.CFUNCTYPE(None, ctypes.c_char_p, ctypes.c_char_p,
                                         ctypes.c_char_p)
    SANE_LIB.sane_init.argtypes = [
        ctypes.POINTER(ctypes.c_int), AUTH_CALLBACK_DEF
    ]
    SANE_LIB.sane_init.restype = ctypes.c_int

    SANE_LIB.sane_exit.argtypes = []
    SANE_LIB.sane_exit.restype = None

    SANE_LIB.sane_get_devices.argtypes = [
        ctypes.POINTER(ctypes.POINTER(ctypes.POINTER(SaneDevice))),
        ctypes.c_int
    ]
    SANE_LIB.sane_get_devices.restype = ctypes.c_int

    SANE_LIB.sane_open.argtypes = [
        ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_void_p),
    ]
    SANE_LIB.sane_open.restype = ctypes.c_int

    SANE_LIB.sane_close.argtypes = [ctypes.c_void_p]
    SANE_LIB.sane_close.restype = None

    SANE_LIB.sane_get_option_descriptor.argtypes = [
        ctypes.c_void_p,
        ctypes.c_int
    ]
    SANE_LIB.sane_get_option_descriptor.restype = \
        ctypes.POINTER(SaneOptionDescriptor)

    SANE_LIB.sane_control_option.argtypes = [
        ctypes.c_void_p,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_void_p,
        ctypes.POINTER(ctypes.c_int)
    ]
    SANE_LIB.sane_control_option.restype = ctypes.c_int

    SANE_LIB.sane_get_parameters.argtypes = [
        ctypes.c_void_p,
        ctypes.POINTER(SaneParameters),
    ]
    SANE_LIB.sane_get_parameters.restype = ctypes.c_int

    SANE_LIB.sane_start.argtypes = [ctypes.c_void_p]
    SANE_LIB.sane_start.restype = ctypes.c_int

    SANE_LIB.sane_read.argtypes = [
        ctypes.c_void_p,
        ctypes.c_void_p,
        ctypes.c_int,
        ctypes.POINTER(ctypes.c_int)
    ]
    SANE_LIB.sane_read.restype = ctypes.c_int

    SANE_LIB.sane_cancel.argtypes = [ctypes.c_void_p]
    SANE_LIB.sane_cancel.restype = None

    SANE_LIB.sane_set_io_mode.argtypes = [ctypes.c_void_p, ctypes.c_int]
    SANE_LIB.sane_set_io_mode.restype = ctypes.c_int

    SANE_LIB.sane_get_select_fd.argtypes = [ctypes.c_void_p,
                                            ctypes.POINTER(ctypes.c_int)]
    SANE_LIB.sane_get_select_fd.restype = ctypes.c_int


def is_sane_available():
    global sane_available
    return sane_available


def sane_init(auth_callback=__dummy_auth_callback):
    global sane_available, sane_is_init, sane_version
    assert(sane_available)

    sane_is_init += 1
    if sane_is_init > 1:
        return sane_version

    version_code = ctypes.c_int()
    wrap_func = __AuthCallbackWrapper(auth_callback).wrapper
    auth_callback = AUTH_CALLBACK_DEF(wrap_func)

    status = SANE_LIB.sane_init(ctypes.pointer(version_code), auth_callback)
    if status != SaneStatus.GOOD:
        raise SaneException(SaneStatus(status))
    version_code = version_code.value
    major = (version_code >> 24) & 0xFF
    minor = (version_code >> 16) & 0xFF
    build = (version_code >> 0) & 0xFFFF
    sane_version = SaneVersion(major, minor, build)
    return sane_version


def sane_exit():
    global sane_available, sane_is_init
    assert(sane_available)

    # TODO(Jflesch): This is a workaround
    # In a multithreaded environment, for some unknown reason,
    # calling sane_exit() will work but the program will crash
    # when stopping. So we simply never call sane_exit() ...

    # sane_is_init -= 1
    # if sane_is_init <= 0:
    #     SANE_LIB.sane_exit()


def sane_get_devices(local_only=False):
    global sane_available
    assert(sane_available)

    devices_ptr = ctypes.POINTER(ctypes.POINTER(SaneDevice))()

    status = SANE_LIB.sane_get_devices(ctypes.pointer(devices_ptr),
                                       ctypes.c_int(local_only))
    if status != SaneStatus.GOOD:
        raise SaneException(SaneStatus(status))

    devices = []
    i = 0
    while devices_ptr[i]:
        devices.append(devices_ptr[i].contents)
        i += 1
    return devices


def sane_open(dev_name):
    global sane_available
    assert(sane_available)

    if isinstance(dev_name, str):
        dev_name = dev_name.encode('utf-8')

    handle_ptr = ctypes.c_void_p()

    status = SANE_LIB.sane_open(ctypes.c_char_p(dev_name),
                                ctypes.pointer(handle_ptr))
    if status != SaneStatus.GOOD:
        raise SaneException(SaneStatus(status))

    return handle_ptr


def sane_close(handle):
    global sane_available
    assert(sane_available)

    SANE_LIB.sane_close(handle)


def sane_get_option_descriptor(handle, option_idx):
    global sane_available
    assert(sane_available)

    opt_desc_ptr = SANE_LIB.sane_get_option_descriptor(
        handle, ctypes.c_int(option_idx))
    if not opt_desc_ptr:
        raise SaneException(SaneStatus(SaneStatus.INVAL))
    return opt_desc_ptr.contents


def sane_get_option_value(handle, option_idx):
    global sane_available
    assert(sane_available)

    # we need the descriptor first in order to allocate a buffer of the correct
    # size, then cast it to the correct type
    opt_desc = sane_get_option_descriptor(handle, option_idx)

    buf = ctypes.c_buffer(max(4, opt_desc.size))
    info = ctypes.c_int()

    status = SANE_LIB.sane_control_option(handle, ctypes.c_int(option_idx),
                                          SaneAction.GET_VALUE,
                                          ctypes.pointer(buf),
                                          ctypes.pointer(info))
    if status != SaneStatus.GOOD:
        raise SaneException(SaneStatus(status))

    return SaneValueType(opt_desc.type).buf_to_pyobj(buf)


def sane_set_option_value(handle, option_idx, new_value):
    global sane_available
    assert(sane_available)

    if isinstance(new_value, str):
        new_value = new_value.encode('utf-8')

    # we need the descriptor first in order to allocate a buffer of the correct
    # size, then cast it to the correct type
    opt_desc = sane_get_option_descriptor(handle, option_idx)

    value = SaneValueType(opt_desc.type).get_ctype_obj(new_value)
    info = ctypes.c_int()

    status = SANE_LIB.sane_control_option(handle, ctypes.c_int(option_idx),
                                          SaneAction.SET_VALUE,
                                          ctypes.pointer(value),
                                          ctypes.pointer(info))
    if status != SaneStatus.GOOD:
        raise SaneException(SaneStatus(status))

    return SaneInfo(info.value)


def sane_set_option_auto(handle, option_idx):
    global sane_available
    assert(sane_available)

    info = ctypes.c_int()

    status = SANE_LIB.sane_control_option(handle, ctypes.c_int(option_idx),
                                          SaneAction.SET_AUTO,
                                          ctypes.c_void_p(),
                                          ctypes.pointer(info))
    if status != SaneStatus.GOOD:
        raise SaneException(SaneStatus(status))

    return SaneInfo(info.value)


def sane_get_parameters(handle):
    global sane_available
    assert(sane_available)

    parameters = SaneParameters()

    status = SANE_LIB.sane_get_parameters(handle, ctypes.pointer(parameters))
    if status != SaneStatus.GOOD:
        raise SaneException(SaneStatus(status))

    return parameters


def sane_start(handle):
    global sane_available
    assert(sane_available)

    status = SANE_LIB.sane_start(handle)
    if status == SaneStatus.NO_DOCS:
        raise StopIteration()
    if status != SaneStatus.GOOD:
        raise SaneException(SaneStatus(status))


def sane_read(handle, nb_bytes):
    global sane_available
    assert(sane_available)

    buf = ctypes.c_buffer(nb_bytes)
    length = ctypes.c_int()

    status = SANE_LIB.sane_read(handle, ctypes.pointer(buf), len(buf),
                                ctypes.pointer(length))
    if status == SaneStatus.NO_DOCS:
        raise StopIteration()
    elif status == SaneStatus.EOF:
        raise EOFError()
    elif status != SaneStatus.GOOD:
        raise SaneException(SaneStatus(status))
    return buf[:length.value]


def sane_cancel(handle):
    global sane_available
    assert(sane_available)

    SANE_LIB.sane_cancel(handle)


def sane_set_io_mode(handle, non_blocking=False):
    global sane_available
    assert(sane_available)

    status = SANE_LIB.sane_set_io_mode(handle,
                                       ctypes.c_int(int(non_blocking)))
    if status != SaneStatus.GOOD:
        raise SaneException(SaneStatus(status))


def sane_get_select_fd(handle):
    global sane_available
    assert(sane_available)

    fd = ctypes.c_int()
    status = SANE_LIB.sane_get_select_fd(handle, ctypes.pointer(fd))
    if status != SaneStatus.GOOD:
        raise SaneException(SaneStatus(status))
    return fd.value


def sane_strstatus(status):
    global sane_available
    assert(sane_available)

    if status in SaneStatus.__meanings:
        return SaneStatus.__meanings[status]
    return "Unknown error code: %d" % (status)



import logging
import os
import sys


logger = logging.getLogger(__name__)


__all__ = [
    'init',
    'exit',
    'Scanner',
    'ScannerOption',
    'PyinsaneException',
    'get_devices',
    'set_scanner_opt',
    'maximize_scan_area',
]

from .util import PyinsaneException

if os.name == "nt":
    from .wia.abstract import *
elif sys.platform == "darwin":
    # The dedicated process appear to crash on MacOSX. Don't know why.
    from .sane.abstract import *
else:
    from .sane.abstract_proc import *


def __normalize_value(value):
    if isinstance(value, str):
        return value.lower()
    return value


def set_scanner_opt(scanner, opt, values):
    """
    Utility function to set most common values easily.

    Examples:
        set_scanner_opt(scanner, "source", ["flatbed"])
        set_scanner_opt(scanner, "source", ["ADF", "feeder"])
    """
    assert(values is not None and values != [])

    if opt not in scanner.options:
        # check it's not just a casing problem
        for key in scanner.options.keys():
            if opt.lower() == key.lower():
                opt = key
                break
        # otherwise just keep going, it will raise a KeyError anyway

    if not scanner.options[opt].capabilities.is_active():
        logger.error(
            "Unable to set scanner option [{}]:"
            " Option is not active".format(opt)
        )
        # this may not be a problem. For instance, 'source' is not active
        # on all scanners, and there is no point in raising an exception
        # when trying to set it to 'FlatBed' or 'Auto'
        return

    last_exc = None
    for value in values:

        # See if we can normalize it first
        if isinstance(scanner.options[opt].constraint, list):
            found = False
            for possible in scanner.options[opt].constraint:
                if __normalize_value(value) == __normalize_value(possible):
                    value = possible
                    found = True
                    break
            if not found:
                # no direct match. See if we have an indirect one
                # for instance, 'feeder' in 'Automatic Document Feeder'
                for possible in scanner.options[opt].constraint:
                    if isinstance(possible, str) and __normalize_value(value) in __normalize_value(possible):
                        logger.info(
                            "Value for [{}] changed from [{}] to [{}]".format(
                                opt, value, possible
                            )
                        )
                        value = possible
                        found = True
                        break
            # beware: don't select a source that is not in the constraint,
            # with some drivers (Brother DCP-8025D for instance),
            # it may segfault.
            if not found:
                last_exc = PyinsaneException(
                    "Invalid value [{}] for option [{}]."
                    " Valid values are [{}]".format(
                        value, opt, scanner.options[opt].constraint
                    )
                )
                continue

        # Then try to set it
        try:
            scanner.options[opt].value = value
            logger.info("[{}] set to [{}]".format(opt, value))
            return
        except (KeyError, PyinsaneException) as exc:
            logger.info("Failed to set [{}] to [{}]: [{}]".format(
                opt, str(value), str(exc))
            )
            last_exc = exc
    logger.warning("Failed to set [{}] to [{}]: [{}]".format(
        opt, values, last_exc)
    )
    raise last_exc


def __set_scan_area_pos(options, opt_name, select_value_func, missing_options):
    if opt_name not in options:
        if missing_options:
            missing_options.append(opt_name)
    else:
        if not options[opt_name].capabilities.is_active():
            logger.warning(
                "Unable to set scanner option [{}]:"
                " Option is not active".format(opt_name)
            )
            return
        constraint = options[opt_name].constraint
        if isinstance(constraint, tuple):
            value = select_value_func(constraint[0], constraint[1])
        elif isinstance(constraint, list):
            value = select_value_func(constraint)
        options[opt_name].value = value


def maximize_scan_area(scanner):
    """
    Utility function to make sure the scanner scan the biggest area possible.
    Must be called *after* setting the resolution.
    """
    opts = scanner.options
    missing_opts = []
    __set_scan_area_pos(opts, "tl-x", min, missing_opts)
    __set_scan_area_pos(opts, "tl-y", min, missing_opts)
    __set_scan_area_pos(opts, "br-x", max, missing_opts)
    __set_scan_area_pos(opts, "br-y", max, missing_opts)
    __set_scan_area_pos(opts, "page-height", max, None)
    __set_scan_area_pos(opts, "page-width", max, None)
    if missing_opts:
        logger.warning(
            "Failed to maximize the scan area. Missing options: {}".format(
                ", ".join(missing_opts)
            )
        )



import logging


logger = logging.getLogger(__name__)


__all__ = [
    'AliasOption',
    'PyinsaneException',
]


class PyinsaneException(Exception):
    def __init__(self, status):
        Exception.__init__(self, str(status))
        self.status = status


class AliasOption(object):
    def __init__(self, name, alias_for, options):
        self.__dict__['alias_for'] = alias_for
        self.__dict__['_options'] = [
            options[opt_name] for opt_name in alias_for
        ]
        self.__dict__['name'] = name

    def __getattr__(self, attr):
        if '_options' not in self.__dict__:
            raise AttributeError()
        return getattr(self.__dict__['_options'][0], attr)

    def __setattr__(self, attr, new_value):
        last_exc = None
        for opt in self.__dict__['_options']:
            try:
                setattr(opt, attr, new_value)
            except Exception as exc:
                # keep trying to set the options for consistency
                # (some driver may return an error while accepting the value
                # ...)
                logger.exception("Failed to set option {}: {}".format(
                    self.__dict__['name'], exc
                ))
                last_exc = exc
        if last_exc:
            raise last_exc

    def __str__(self):
        return ("Option [{}] (alias for {})".format(
            self.__dict__['name'], self.__dict__['alias_for']
        ))



import io
import logging

import PIL.Image
import PIL.ImageFile

from . import rawapi
from .. import util
from .rawapi import WIAException


__all__ = [
    'init',
    'exit',
    'Scanner',
    'ScannerOption',
    'WIAException',
    'get_devices',
]


logger = logging.getLogger(__name__)


def init():
    rawapi.init()


def exit():
    rawapi.exit()


class Scan(object):

    # WORKAROUND(Jflesch)> When using the ADF with HP drivers, even if there is
    # no paper in the ADF
    # The driver will write about 56 bytes in the stream (BMP headers ?)
    # --> We ignore BMP too small
    MIN_BYTES = 1024

    def __init__(self, session, source, multiple=False):
        self._session = session
        self.source = source
        self._data = b""
        self._img_size = None
        self.scan = rawapi.start_scan(self.source)
        self.multiple = multiple

    def read(self):
        # will raise EOFError at the end of each page
        # will raise StopIteration when all the pages are done
        try:
            buf = self.scan.read()
            self._data += buf
            self._got_data = True
        except EOFError:
            if len(self._data) >= self.MIN_BYTES:
                self._session._add_image(self._get_current_image())
                if self.multiple:
                    self._session._next()
                raise
            else:
                # Too small. Scrap the crap from the drivers.
                self._data = b""
                raise StopIteration()

    def _get_current_image(self):
        # We get the image as a truncated bitmap.
        # ('rawrgb' is not supported by all drivers ...)
        # Bitmap headers are annoying.
        PIL.ImageFile.LOAD_TRUNCATED_IMAGES = True
        stream = io.BytesIO(self._data)
        img = PIL.Image.open(stream)
        self._img_size = img.size
        return img

    def _get_available_lines(self):
        if self._img_size is None:
            try:
                self._get_current_image()
            except:
                # assumes we just got truncated headers for now
                return (0, 0)
        # estimated
        line_size = self._img_size[0] * 3  # rgb
        data = len(self._data) - 1024  # - headers
        return (0, int(data / line_size))

    available_lines = property(_get_available_lines)

    def _get_expected_size(self):
        if self._img_size:
            return self._img_size
        options = self._session.scanner.options
        return (
            int(options['xextent'].value),
            int(options['yextent'].value)
        )

    expected_size = property(_get_expected_size)

    def get_image(self, start_line=0, end_line=-1):
        img = self._get_current_image()
        img_size = img.size
        if start_line > 0:
            img = img.crop((0, start_line, img_size[0], img_size[1]))
        if end_line >= 0:
            end_line -= start_line
            img = img.crop((0, 0, img_size[0], end_line))
        return img

    def cancel(self):
        # TODO
        raise NotImplementedError()

    def __str__(self):
        return ("Scan instance for session {}".format(self._session))


class ScanSession(object):
    def __init__(self, scanner, srcid, multiple):
        self.scanner = scanner
        self.multiple = multiple
        self.source = scanner.srcs[srcid]
        self.images = []
        self.scan = Scan(self, self.source, self.multiple)

    def _add_image(self, img):
        self.images.append(img)

    def _next(self):
        self.scan = Scan(self, self.source, self.multiple)


class ScannerCapabilities(object):
    def __init__(self, opt):
        self.opt = opt

    def is_active(self):
        return True

    def is_settable(self):
        return self.opt.accessright == "rw"

    def __str__(self):
        return ("Access: {}".format(self.opt.accessright))


class ScannerOption(object):
    idx = -1
    title = ""
    desc = ""
    val_type = None  # TODO
    unit = None  # TODO
    size = 4
    capabilities = None

    constraint_type = None  # TODO
    constraint = None

    def __init__(self, scanner, objsrc, name, value, possible_values,
                 accessright, constraint):
        self.objsrc = objsrc
        self.scanner = scanner
        self.name = name
        self._value = value
        if constraint:
            self.constraint = constraint
        else:
            self.constraint = possible_values
        self.accessright = accessright
        self.capabilities = ScannerCapabilities(self)

    def _get_value(self):
        return self._value

    def _set_value(self, new_value):
        if self.accessright != 'rw':
            raise rawapi.WIAException("Property {} is read-only".format(
                self.name
            ))
        has_success = False
        exc = None
        for obj in self.objsrc:
            try:
                rawapi.set_property(obj, self.name, new_value)
                has_success = True
            except rawapi.WIAException as _exc:
                logger.warning("Exception while setting {}: {}".format(
                    self.name, _exc
                ))
                logger.exception(_exc)
                exc = _exc
        self._value = new_value
        try:
            self.scanner.reload_options()
        except:
            pass
        if not has_success:
            raise exc

    value = property(_get_value, _set_value)

    def __str__(self):
        return ("Option [{}] (basic)".format(self.name))

    def __eq__(self, other):
        return (self.name == other.name and self.constraint == other.constraint)


class SourceOption(ScannerOption):
    idx = -1
    title = ""
    desc = ""
    val_type = None  # TODO
    unit = None  # TODO
    size = 4
    capabilities = None

    constraint_type = None  # TODO
    constraint = None

    accessright = "rw"

    def __init__(self, sources):
        self.name = "source"
        self.sources = sources
        self.constraint = [srcid for (srcid, src) in sources]
        self._value = self.constraint[0]
        self.capabilities = ScannerCapabilities(self)

    def _get_value(self):
        return self._value

    def _set_value(self, new_value):
        if new_value not in self.constraint:
            raise WIAException("Invalid source: {}".format(new_value))
        self._value = new_value

    value = property(_get_value, _set_value)

    def __str__(self):
        return ("Option [{}] (source)".format(self.name))

    def __eq__(self, other):
        return isinstance(other, SourceOption)


class ModeOption(object):
    idx = -1
    title = ""
    desc = ""
    val_type = None  # TODO
    unit = None  # TODO
    size = 4
    capabilities = None

    constraint_type = None  # TODO
    constraint = None

    def __init__(self, scanner):
        self.name = "mode"
        self.scanner = scanner
        self.constraint = ["Color", "Gray", "BW"]
        self.capabilities = ScannerCapabilities(self)
        self.scanner = scanner

    def _get_value(self):
        opts = self.scanner.options
        if opts['bits_per_channel'].value == 1:
            return 'BW'
        if opts['channels_per_pixel'].value == 1:
            return 'Gray'
        return 'Color'

    def _set_value(self, new_value):
        opts = self.scanner.options
        if new_value == "BW":
            opts['depth'].value = 1
            self.scanner.reload_options()
            return
        if new_value == "Gray":
            opts['depth'].value = 8
            self.scanner.reload_options()
            return
        if new_value == "Color":
            opts['depth'].value = 24
            self.scanner.reload_options()
            return
        raise WIAException("Unknown value '{}' for option 'mode'".format(
            new_value
        ))

    value = property(_get_value, _set_value)

    def _get_accessright(self):
        opts = self.scanner.options
        return opts['depth'].accessright

    accessright = property(_get_accessright)

    def __str__(self):
        return ("Option [{}] (mode)".format(self.name))


def get_pos_constraint(options, opt_min, opt_max, opt_res):
    if (opt_min not in options or
            opt_max not in options or
            opt_res not in options):
        return None
    vmax = options[opt_max].value / 1000  # thousandths of inch
    vres = options[opt_res].value
    return (0, int(vmax * vres))


class PosOption(object):
    idx = -1
    title = ""
    desc = ""
    val_type = None  # TODO
    unit = None  # TODO
    size = 4
    capabilities = None

    constraint_type = None  # TODO
    constraint = None

    def __init__(self, scanner, name, base_name, options, opt_min, opt_max,
                 opt_res):
        self.name = name
        self.base_name = base_name
        self.capabilities = ScannerCapabilities(self)
        self.scanner = scanner
        self._options = options
        self.constraint = get_pos_constraint(options, opt_min, opt_max, opt_res)

    def _get_value(self):
        return self._options[self.base_name + 'pos'].value

    def _set_value(self, new_value):
        diff = self._options[self.base_name + 'pos'].value - self.value
        extent = self._options[self.base_name + 'extent'].value
        self._options[self.base_name + 'pos'].value = new_value
        self._options[self.base_name + 'extent'].value = extent - diff

    value = property(_get_value, _set_value)

    def _get_accessright(self):
        rw = True
        if self._options[self.base_name + 'pos'].accessright != 'rw':
            rw = False
        elif self._options[self.base_name + 'extent'].accessright != 'rw':
            rw = False
        return "rw" if rw else "ro"

    accessright = property(_get_accessright)

    def __str__(self):
        return ("Option [{}] (position for [{}])".format(
            self.name, self.base_name
        ))


class ExtendOption(object):
    idx = -1
    title = ""
    desc = ""
    val_type = None  # TODO
    unit = None  # TODO
    size = 4
    capabilities = None

    constraint_type = None  # TODO
    constraint = None

    def __init__(self, scanner, name, base_name, options, opt_min, opt_max,
                 opt_res):
        self.name = name
        self.base_name = base_name
        self.capabilities = ScannerCapabilities(self)
        self.scanner = scanner
        self._options = options
        self.constraint = get_pos_constraint(options, opt_min, opt_max, opt_res)

    def _get_value(self):
        return (self._options[self.base_name + 'extent'].value +
                self._options[self.base_name + 'pos'].value)

    def _set_value(self, new_value):
        new_value -= self._options[self.base_name + 'pos'].value
        self._options[self.base_name + 'extent'].value = new_value

    value = property(_get_value, _set_value)

    def _get_accessright(self):
        rw = True
        if self._options[self.base_name + 'pos'].accessright != 'rw':
            rw = False
        elif self._options[self.base_name + 'extent'].accessright != 'rw':
            rw = False
        return "rw" if rw else "ro"

    accessright = property(_get_accessright)

    def __str__(self):
        return ("Option [{}] (extent for [{}])".format(
            self.name, self.base_name
        ))


class Scanner(object):
    def __init__(self, name):
        self._dev = rawapi.open(name)
        self._srcs_list = rawapi.get_sources(self._dev)
        self.srcs = {}
        for (srcid, src) in self._srcs_list:
            self.srcs[srcid] = src

        self.options = {}
        self.reload_options()

        self.name = name
        self.nice_name = self.options['dev_name'].value
        self.vendor = self.options['vend_desc'].value
        self.model = self.options['dev_desc'].value
        self.dev_type = self.options['dev_type'].value

        for (opt, val) in [
            ('current_intent', 'image_type_color,maximize_quality'),
            ('format', 'bmp'),
            ('preferred_format', 'bmp'),
            ('page_size', 'a4'),
            ('depth', 24),
        ]:
            if opt not in self.options:
                continue
            try:
                self.options[opt].value = val
                logger.warning("Option '{}' preset to '{}' on [{}]".format(opt, val, self))
            except:
                logger.warning("Failed to pre-set option '{}' on [{}]".format(opt, self))

    @staticmethod
    def _convert_prop_list_to_dict(props):
        out = {}
        for (propname, propvalue, accessright, possible_values) in props:
            out[propname] = {
                'value': propvalue,
                'accessright': accessright,
                'possible_values': possible_values
            }
        return out

    @staticmethod
    def _merge_constraints(props, constraints):
        for (propname, constraint) in constraints:
            if propname not in props:
                logger.warning("Constraint found on property [{}] but property not found".format(propname))
                continue
            if isinstance(constraint, list):
                constraint.sort()
            props[propname]['constraint'] = constraint

    def reload_options(self):
        original = self.options

        self.options = {}

        dev_properties = self._convert_prop_list_to_dict(
            rawapi.get_properties(self._dev)
        )
        dev_constraints = rawapi.get_constraints(self._dev)
        self._merge_constraints(dev_properties, dev_constraints)

        src_properties = {}
        for (srcid, src) in self._srcs_list:
            src_properties[srcid] = self._convert_prop_list_to_dict(
                rawapi.get_properties(src)
            )
            src_constraints = rawapi.get_constraints(src)
            self._merge_constraints(src_properties[srcid], src_constraints)

        for (opt_name, opt_infos) in dev_properties.items():
            self.options[opt_name] = ScannerOption(
                self,
                [self._dev],
                opt_name, opt_infos['value'], opt_infos['possible_values'],
                opt_infos['accessright'],
                opt_infos['constraint'] if 'constraint' in opt_infos else None
            )
        # generate list of options from all the sources, and try to apply them
        # on all the sources
        for (srcid, opts) in src_properties.items():
            for (opt_name, opt_infos) in opts.items():
                opt = ScannerOption(
                    self,
                    self.srcs.values(),
                    opt_name, opt_infos['value'], opt_infos['possible_values'],
                    opt_infos['accessright'],
                    opt_infos['constraint'] if 'constraint' in opt_infos else None
                )
                if opt_name in self.options:
                    if self.options[opt_name] != opt:
                        logger.warning("Got multiple time the option [{}], but they are not identical".format(opt_name))
                self.options[opt_name] = opt

        # aliases to match Sane
        if "xpos" in self.options.keys() and "xextent" in self.options.keys():
            self.options['tl-x'] = PosOption(
                self, "tl-x", "x", self.options, "min_horizontal_size",
                "max_horizontal_size", "xres"
            )
            self.options['br-x'] = ExtendOption(
                self, "br-x", "x", self.options, "min_horizontal_size",
                "max_horizontal_size", "xres"
            )
        if "ypos" in self.options.keys() and "yextent" in self.options.keys():
            self.options['tl-y'] = PosOption(
                self, "tl-y", "y", self.options, "min_vertical_size",
                "max_vertical_size", "yres"
            )
            self.options['br-y'] = ExtendOption(
                self, "br-y", "y", self.options, "min_vertical_size",
                "max_vertical_size", "yres"
            )
        res_alias_for = []
        if "xres" in self.options.keys():
            res_alias_for.append("xres")
        if "yres" in self.options.keys():
            res_alias_for.append("yres")
        if res_alias_for != []:
            self.options['resolution'] = util.AliasOption(
                "resolution", res_alias_for, self.options
            )

        if 'source' in original:
            self.options['source'] = original['source']
        else:
            self.options['source'] = SourceOption(self._srcs_list)
        if 'mode' in original:
            self.options['mode'] = original['mode']
        else:
            self.options['mode'] = ModeOption(self)

    def scan(self, multiple=False):
        if 'pages' in self.options:
            try:
                # Even with an ADF, Pyinsane actually request one page
                # after the other.
                # This is not orthodox at all, but still, it has proven to be
                # the most reliable way.
                self.options['pages'].value = 1
            except:
                logger.exception("Failed to set options [pages]")
        return ScanSession(self, self.options['source'].value, multiple)

    def __str__(self):
        return ("'%s' (%s, %s, %s)"
                % (self.nice_name, self.vendor, self.model, self.dev_type))


def get_devices(local_only=False):
    devs = rawapi.get_devices()
    out = []
    for dev in devs:
        try:
            scanner = Scanner(dev[0])
            out.append(scanner)
        except Exception as exc:
            logger.warning("Failed to access scanner {} : {}".format(dev, exc))
            logger.exception(exc)
    return out



#!/usr/bin/env python3

import pyinsane2


if __name__ == "__main__":
    pyinsane2.init()
    try:
        for device in pyinsane2.get_devices():
            print("%s" % (str(device)))

            for opt in device.options.values():
                print("  Option: %s" % (opt.name))
                print("    Title: %s" % (opt.title))
                print("    Desc: %s" % (opt.desc))
                print("    Type: %s" % (str(opt.val_type)))
                print("    Unit: %s" % (str(opt.unit)))
                print("    Size: %d" % (opt.size))
                print("    Capabilities: %s" % (str(opt.capabilities)))
                print("    Constraint type: %s" % (str(opt.constraint_type)))
                print("    Constraint: %s" % (str(opt.constraint)))
                try:
                    print("    Value: %s" % (str(opt.value)))
                except pyinsane2.PyinsaneException as exc:
                    # Some scanner allow changing a value, but not reading it.
                    # For instance Canon Lide 110 allow setting the resolution,
                    # but not reading it
                    print("    Value: Failed to get the value: %s" % str(exc))

            print("")
    finally:
        pyinsane2.exit()



#!/usr/bin/env python3

import os
import sys

import pyinsane2


def main(args):
    dstdir = args[0]

    devices = pyinsane2.get_devices()
    assert(len(devices) > 0)
    device = devices[0]

    print("Will use the scanner [%s](%s)"
          % (str(device), device.name))

    pyinsane2.set_scanner_opt(device, "source", ["ADF", "Feeder"])
    # Beware: Some scanner have "Lineart" or "Gray" as default mode
    pyinsane2.set_scanner_opt(device, "mode", ["Color"])
    pyinsane2.set_scanner_opt(device, "resolution", [300])
    pyinsane2.maximize_scan_area(device)

    # Note: If there is no page in the feeder, the behavior of device.scan()
    # is not guaranteed : It may raise StopIteration() immediately
    # or it may raise it when scan.read() is called

    try:
        scan_session = device.scan(multiple=True)
        print("Scanning ...")
        while True:
            try:
                scan_session.scan.read()
            except EOFError:
                print("Got page %d" % (len(scan_session.images)))
                img = scan_session.images[-1]
                imgpath = os.path.join(dstdir, "%d.jpg" %
                                       (len(scan_session.images)))
                img.save(imgpath)
    except StopIteration:
        print("Got %d pages" % len(scan_session.images))


if __name__ == "__main__":
    args = sys.argv[1:]
    if len(args) <= 0 or args[0][0] == '-':
        print("Usage:")
        print("  %s <dst directory>" % sys.argv[0])
        sys.exit(1)
    pyinsane2.init()
    try:
        main(args)
    finally:
        pyinsane2.exit()



import sys

from PIL import Image

from . import rawapi
from .. import util

# import basic elements directly, so the caller
# doesn't have to import rawapi if they need them.
from .rawapi import SaneCapabilities
from .rawapi import SaneConstraint
from .rawapi import SaneConstraintType
from .rawapi import SaneException
from .rawapi import SaneStatus
from .rawapi import SaneUnit
from .rawapi import SaneValueType
from .rawapi import sane_init, sane_exit


__all__ = [
    'SaneCapabilities',
    'SaneConstraint',
    'SaneConstraintType',
    'SaneException',
    'SaneStatus',
    'SaneValueType',
    'SaneUnit',
    'init',
    'exit',
    'Scanner',
    'ScannerOption',
    'get_devices',
]

# We use huge buffers to spend the maximum amount of time in non-Python code
SANE_READ_BUFSIZE = 512 * 1024

# XXX(Jflesch): Never open more than one handle at the same time.
# Some Sane backends don't support it. For instance, I have 2 HP scanners, and
# if I try to access both from the same process, I get I/O errors.
sane_dev_handle = ("", None)


def init():
    sane_init()


def exit():
    sane_exit()


class ScannerOption(object):
    idx = 0
    name = ""
    title = ""
    desc = ""
    val_type = SaneValueType(SaneValueType.INT)
    unit = SaneUnit(SaneUnit.NONE)
    size = 4
    capabilities = SaneCapabilities(SaneCapabilities.NONE)

    constraint_type = SaneConstraintType(SaneConstraintType.NONE)
    constraint = None

    def __init__(self, scanner, idx):
        self.__scanner = scanner
        self.idx = idx

    @staticmethod
    def build_from_rawapi(scanner, opt_idx, opt_raw):
        opt = ScannerOption(scanner, opt_idx)
        opt.name = opt_raw.name
        if opt.name is not None and hasattr(opt.name, "decode"):
            opt.name = opt.name.decode('utf-8')
        opt.title = opt_raw.title
        if opt.title is not None and hasattr(opt.title, "decode"):
            opt.title = opt.title.decode('utf-8')
        opt.desc = opt_raw.desc
        if opt.desc is not None and hasattr(opt.desc, "decode"):
            opt.desc = opt.desc.decode('utf-8')  # TODO : multi-line
        opt.val_type = SaneValueType(opt_raw.type)
        opt.unit = SaneUnit(opt_raw.unit)
        opt.size = opt_raw.size
        opt.capabilities = SaneCapabilities(opt_raw.cap)
        opt.constraint_type = SaneConstraintType(opt_raw.constraint_type)
        opt.constraint = opt.constraint_type.get_pyobj_constraint(
            opt_raw.constraint)
        return opt

    def _get_value(self):
        self.__scanner._open()
        val = rawapi.sane_get_option_value(sane_dev_handle[1], self.idx)
        if not self.capabilities.is_active():
            # XXX(Jflesch): if the option is not active, some backends still
            # return a value, some don't and return an error instead.
            # To avoid mistakes in user programs, we make the behavior here
            # consistent and always raise an exception.
            raise SaneException("Option '%s' is not active" % self.name)
        if hasattr(val, 'decode'):
            val = val.decode("utf-8")
        return val

    def _set_value(self, new_value):
        self.__scanner._open()
        rawapi.sane_set_option_value(sane_dev_handle[1], self.idx, new_value)

    value = property(_get_value, _set_value)


class ImgUtil(object):
    COLOR_BYTES = {
        1: {  # we expanded the bits to bytes on-the-fly
            "L": 1,
            "RGB": 3,
        },
        8: {
            "L": 1,
            "RGB": 3,
        },
        16: {
            "L": 2,
            "RGB": 6
        }
    }

    @staticmethod
    def unpack_1_to_8(whole_raw_packed, pixels_per_line, bytes_per_line):
        # Each color is on one bit. We unpack immediately so each color
        # is on one byte.
        # We must take care of one thing : the last byte of each line
        # contains unused bits. They must be dropped
        whole_raw_unpacked = b""

        if sys.version_info < (3, ):
            positive_bit = chr(0x00)
            negative_bit = chr(0xFF)
        else:
            positive_bit = bytes([0x00])
            negative_bit = bytes([0xFF])

        for chunk in range(0, len(whole_raw_packed), bytes_per_line):
            raw_packed = whole_raw_packed[:bytes_per_line]
            whole_raw_packed = whole_raw_packed[bytes_per_line:]
            raw_unpacked = b""

            for byte in raw_packed:
                if type(byte) == str:
                    byte = ord(byte)
                for bit in range(7, -1, -1):
                    if ((byte & (1 << bit)) > 0):
                        raw_unpacked += positive_bit
                    else:
                        raw_unpacked += negative_bit
            assert(len(raw_packed) * 8 == len(raw_unpacked))

            raw_unpacked = raw_unpacked[:pixels_per_line]
            whole_raw_unpacked += raw_unpacked

        return whole_raw_unpacked

    @staticmethod
    def raw_to_img(raw, parameters):
        mode = rawapi.SaneFrame(parameters.format).get_pil_format()
        # color_bytes = ImgUtil.COLOR_BYTES[parameters.depth][mode]
        width = parameters.pixels_per_line
        height = (len(raw) / parameters.bytes_per_line)
        if parameters.depth == 1:
            raw = ImgUtil.unpack_1_to_8(raw, width,
                                        parameters.bytes_per_line)
        return Image.frombuffer(mode, (int(width), int(height)), raw, "raw",
                                mode, 0, 1)


class Scan(object):
    def __init__(self, scanner):
        self.scanner = scanner
        self.__session = None
        self.__raw_lines = []
        self.__img_finished = False

    def _set_session(self, session):
        self.__session = session

    def _init(self):
        self.scanner._open()
        rawapi.sane_start(sane_dev_handle[1])
        try:
            self.parameters = \
                rawapi.sane_get_parameters(sane_dev_handle[1])
        except Exception:
            rawapi.sane_cancel(sane_dev_handle[1])
            raise

    def read(self):
        if self.__img_finished:
            # start a new one
            self.__raw_lines = []
            self.__img_finished = False

        try:
            read = rawapi.sane_read(sane_dev_handle[1], SANE_READ_BUFSIZE)
        except EOFError:
            line_size = self.parameters.bytes_per_line
            for line in self.__raw_lines:
                if len(line) != line_size:
                    print(("Pyinsane: Warning: Unexpected line size: %d"
                           " instead of %d") % (len(line), line_size))
            raw = (b'').join(self.__raw_lines)
            # don't do purge the lines here. wait for the next call to read()
            # because, in the meantime, the caller might use get_image()
            self.__img_finished = True
            self.__session.images.append(ImgUtil.raw_to_img(
                raw, self.parameters))
            raise

        # cut what we just read, line by line

        line_size = self.parameters.bytes_per_line

        if (len(self.__raw_lines) > 0):
            cut = line_size - len(self.__raw_lines[-1])
            self.__raw_lines[-1] += read[:cut]
            read = read[cut:]

        range_func = range
        if sys.version_info.major < 3:
            range_func = xrange

        for _ in range_func(0, len(read), line_size):
            self.__raw_lines.append(read[:line_size])
            read = read[line_size:]

        if len(read) > 0:
            self.__raw_lines.append(read)

    def _get_available_lines(self):
        line_size = self.parameters.bytes_per_line
        r = len(self.__raw_lines)
        if (r > 0 and len(self.__raw_lines[-1]) < line_size):
            r -= 1
        return (0, r)

    available_lines = property(_get_available_lines)

    def _get_expected_size(self):
        """
        Returns the expected size of the image (tuple: (w, h)).
        Note that (afaik) Sane makes it mandatory for the driver
        to indicates the length of the lines. However, it is not mandatory
        for the driver to indicates the expected number of lines (for
        instance, hand-held scanners can't know it before the end of
        the scan). In that case, the expected height of the image
        will -1 here.
        """
        width = self.parameters.pixels_per_line
        height = self.parameters.lines
        return (width, height)

    expected_size = property(_get_expected_size)

    def get_image(self, start_line=0, end_line=-1):
        if end_line < 0:
            end_line = len(self.__raw_lines)
        assert(end_line > start_line)
        lines = self.__raw_lines[start_line:end_line]
        lines = b"".join(lines)
        return ImgUtil.raw_to_img(lines, self.parameters)

    def _cancel(self):
        rawapi.sane_cancel(sane_dev_handle[1])


class SingleScan(Scan):
    def __init__(self, scanner):
        Scan.__init__(self, scanner)

        self.is_scanning = True

        self._init()

    def read(self):
        try:
            Scan.read(self)
        except (EOFError, StopIteration):
            self._cancel()
            self.is_scanning = False
            raise

    def cancel(self):
        if self.is_scanning:
            self._cancel()
            self.is_scanning = False


class MultipleScan(Scan):
    def __init__(self, scanner):
        Scan.__init__(self, scanner)
        self.is_scanning = False
        self.is_finished = False
        self.must_request_next_frame = False
        self._init()

    def read(self):
        if self.is_finished:
            raise StopIteration()

        if not self.is_scanning:
            self.is_scanning = True
            self.must_request_next_frame = False

        if self.must_request_next_frame:
            try:
                rawapi.sane_start(sane_dev_handle[1])
            except StopIteration:
                self._cancel()
                self.is_finished = True
                self.is_scanning = False
                raise
            self.must_request_next_frame = False

        try:
            Scan.read(self)
        except EOFError:
            self.must_request_next_frame = True
            raise
        except StopIteration:
            self._cancel()
            self.is_finished = True
            self.is_scanning = False
            # signal the last page first
            raise EOFError()

    def cancel(self):
        if self.is_scanning:
            self._cancel()
            self.is_finished = True
            self.is_scanning = False


class ScanSession(object):
    def __init__(self, scan):
        self.images = []
        self.scan = scan
        self.scan._set_session(self)

    def read(self):
        """
        Deprecated. Use scan_session.scan.read()
        """
        return self.scan.read()

    def get_nb_img(self):
        """
        Deprecated. Use len(scan_session.images) directly
        """
        return len(self.images)

    def get_img(self, idx=0):
        """
        Deprecated. Use scan_session.images[idx] directly
        """
        return self.images[idx]


class Scanner(object):
    def __init__(self, name, vendor="Unknown", model="Unknown",
                 dev_type="Unknown"):
        if hasattr(name, "decode"):
            name = name.decode('utf-8')
        if hasattr(vendor, "decode"):
            vendor = vendor.decode('utf-8')
        if hasattr(model, "decode"):
            model = model.decode('utf-8')
        if hasattr(dev_type, "decode"):
            dev_type = dev_type.decode('utf-8')

        self.name = name
        self.vendor = vendor
        self.model = model
        self.dev_type = dev_type
        self.__options = None  # { "name" : ScannerOption }

    @staticmethod
    def build_from_rawapi(sane_device):
        return Scanner(sane_device.name, sane_device.vendor, sane_device.model,
                       sane_device.type)

    def _open(self):
        global sane_dev_handle
        (devid, handle) = sane_dev_handle
        if devid == self.name:
            return
        self._force_close()
        sane_init()
        handle = rawapi.sane_open(self.name)
        sane_dev_handle = (self.name, handle)

    def _force_close(self):
        global sane_dev_handle
        (devid, handle) = sane_dev_handle
        if handle is None:
            return
        rawapi.sane_close(handle)
        sane_exit()
        sane_dev_handle = ("", None)

    def __load_options(self):
        if self.__options is not None:
            return
        self._open()
        nb_options = rawapi.sane_get_option_value(sane_dev_handle[1], 0)
        self.__options = {}
        for opt_idx in range(1, nb_options):
            opt_desc = rawapi.sane_get_option_descriptor(
                sane_dev_handle[1], opt_idx)
            if not SaneValueType(opt_desc.type).can_getset_opt():
                continue
            opt = ScannerOption.build_from_rawapi(self, opt_idx, opt_desc)
            self.__options[opt.name] = opt

        # WORKAROUND(Jflesch):
        # Lexmark MFP CX510de: option 'resolution' has been mistakenly named
        # 'scan-resolution'
        if ('scan-resolution' in self.__options and
                'resolution' not in self.__options):
            self.__options['resolution'] = util.AliasOption(
                'resolution', ['scan-resolution'], self.__options
            )

        # WORKAROUND(Jflesch):
        # Samsung M288x: option 'source' is actually called 'doc-source'
        if ('doc-source' in self.__options and
                'source' not in self.__options):
            self.__options['source'] = util.AliasOption(
                'source', ['doc-source'], self.__options
            )

    def _get_options(self):
        self.__load_options()
        return self.__options

    options = property(_get_options)

    def scan(self, multiple=False):
        if (not ('source' in self.options and
                 self.options['source'].capabilities.is_active())):
            value = ""
        else:
            value = self.options['source'].value
        if hasattr(value, 'decode'):
            value = value.decode('utf-8')
        if (not multiple or
                ("adf" not in value.lower() and "feeder" not in value.lower())):
            # XXX(Jflesch): We cannot use MultipleScan() with something
            # else than an ADF. If we try, we will never get
            # SANE_STATUS_NO_DOCS from sane_start()/sane_read() and we will
            # loop forever
            scan = SingleScan(self)
        else:
            scan = MultipleScan(self)
        return ScanSession(scan)

    def __str__(self):
        return ("Scanner '%s' (%s, %s, %s)"
                % (self.name, self.vendor, self.model, self.dev_type))


def get_devices(local_only=False):
    sane_init()
    try:
        return [Scanner.build_from_rawapi(device)
                for device in rawapi.sane_get_devices(local_only)]
    finally:
        sane_exit()



#!/usr/bin/env python3

import logging
import os
import pickle
import struct
import sys

import pyinsane2.sane.abstract as pyinsane


logger = logging.getLogger(__name__)


device_cache = {}
scan_sessions = {}


def get_devices(local_only):
    global device_cache

    devices = pyinsane.get_devices(local_only)
    device_cache = {}
    for device in devices:
        device_cache[device.name] = device
    return devices


def get_device(scanner_name):
    global device_cache
    if scanner_name in device_cache:
        return device_cache[scanner_name]
    scanner = pyinsane.Scanner(scanner_name)
    device_cache[scanner_name] = scanner
    return scanner


def get_options(scanner_name):
    return get_device(scanner_name).options


def get_option_value(scanner_name, option_name):
    return get_device(scanner_name).options[option_name].value


def set_option_value(scanner_name, option_name, option_value):
    get_device(scanner_name).options[option_name].value = option_value


def make_scan_session(scanner_name, multiple=False):
    global scan_sessions

    scan_session = get_device(scanner_name).scan(multiple)
    scan_sessions[scanner_name] = scan_session
    return scan_session


def get_images(scanner_name):
    global scan_sessions
    imgs = scan_sessions[scanner_name].images
    imgs = [(img.mode, img.size, img.tobytes()) for img in imgs]
    return imgs


def scan_read(scanner_name):
    global scan_sessions
    return scan_sessions[scanner_name].scan.read()


def get_available_lines(scanner_name):
    global scan_sessions
    return scan_sessions[scanner_name].scan.available_lines


def get_expected_size(scanner_name):
    global scan_sessions
    return scan_sessions[scanner_name].scan.expected_size


def get_image(scanner_name, start_line, end_line):
    global scan_sessions
    img = scan_sessions[scanner_name].scan.get_image(start_line, end_line)
    return (img.mode, img.size, img.tobytes())


def cancel(scanner_name):
    global scan_sessions
    return scan_sessions[scanner_name].scan.cancel()


def exit():
    pass


COMMANDS = {
    "get_devices": get_devices,
    "get_options": get_options,
    "get_option_value": get_option_value,
    "set_option_value": set_option_value,
    "scan": make_scan_session,
    "get_images": get_images,
    "scan_read": scan_read,
    "scan_get_available_lines": get_available_lines,
    "scan_get_expected_size": get_expected_size,
    "scan_get_image": get_image,
    "scan_cancel": cancel,
    "exit": exit,
}


def main_loop(fifo_dir, fifo_filepaths):
    global COMMANDS

    pyinsane.init()

    length_size = len(struct.pack("i", 0))
    fifo_c2s = os.open(fifo_filepaths[0], os.O_RDONLY)
    fifo_s2c = os.open(fifo_filepaths[1], os.O_WRONLY)

    try:
        logger.info("Ready")

        while True:
            length = os.read(fifo_c2s, length_size)
            if length == b'':
                break
            length = struct.unpack("i", length)[0]
            cmd = os.read(fifo_c2s, length)
            if cmd == b'':
                break
            assert(len(cmd) == length)
            cmd = pickle.loads(cmd)

            logger.debug("> {}".format(cmd['command']))
            f = COMMANDS[cmd['command']]
            result = {}
            try:
                result['out'] = f(*cmd['args'], **cmd['kwargs'])
            except BaseException as exc:
                result['exception'] = str(exc.__class__.__name__)
                result['exception_args'] = exc.args
            logger.debug("< {}".format(result))

            result = pickle.dumps(result)
            length = len(result)
            length = struct.pack("i", length)
            os.write(fifo_s2c, length)
            os.write(fifo_s2c, result)

            if cmd['command'] == 'exit':
                break
    finally:
        os.close(fifo_s2c)
        os.close(fifo_c2s)

    logger.info("Daemon stopped")


if __name__ == "__main__":
    main_loop(sys.argv[1], sys.argv[2:4])



#!/usr/bin/env python

import os
import platform

from setuptools import Extension
from setuptools import setup, find_packages

if platform.architecture()[0] == '32bit':
    DEFAULT_ATL_WINDDK_LIB_DIR = "c:\\winddk\\7600.16385.1\\lib\\ATL\\amd64"
else:
    DEFAULT_ATL_WINDDK_LIB_DIR = "c:\\winddk\\7600.16385.1\\lib\\ATL\\i386"

if os.name == "nt":
    extensions = [
        Extension(
            'pyinsane2.wia._rawapi', [
                'pyinsane2/wia/properties.cpp',
                'pyinsane2/wia/rawapi.cpp',
                'pyinsane2/wia/transfer.cpp',
            ],
            include_dirs=[
                # Yeah, I know.
                os.getenv(
                    "WINDDK_INCLUDE_DIR",
                    "c:\\winddk\\7600.16385.1\\inc\\atl71"
                ),
            ],
            library_dirs=[
                # Yeah, I know.
                os.getenv("WINDDK_LIB_DIR", DEFAULT_ATL_WINDDK_LIB_DIR),
            ],
            libraries=[
                "ole32",
                "wiaguid",
            ],
            # extra_compile_args=['/W4'],
            undef_macros=['NDEBUG'],
        ),
    ]
else:
    extensions = []

setup(
    name="pyinsane2",
    # If you change this version, please:
    # - update the ChangeLog file
    version="2.0.10",
    description=(
        "Python library to access and use image scanners (Linux/Windows/etc)"
    ),
    long_description=(
        "Python library to access and use image scanners (devices)."
        " Works on GNU/Linux (Sane), *BSD (Sane), Windows >= Vista (WIA 2),"
        " MacOSX (Sane), etc."
    ),
    keywords="sane scanner",
    url="https://github.com/jflesch/pyinsane",
    download_url="https://github.com/jflesch/pyinsane/archive/2.0.2.zip",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later"
        " (GPLv3+)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Multimedia :: Graphics :: Capture :: Scanners",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    license="GPLv3+",
    author="Jerome Flesch",
    author_email="jflesch@openpaper.work",
    packages=find_packages(),
    data_files=[],
    scripts=[],
    install_requires=[
        "Pillow",
    ],
    ext_modules=extensions,
    setup_requires=['nose>=1.0'],
)



import os
import unittest

import pyinsane2


class TestSaneGetDevices(unittest.TestCase):
    module = None

    def setUp(self):
        pyinsane2.init()

    def test_get_devices(self):
        devices = pyinsane2.get_devices()
        if len(devices) == 0:
            # if there are no devices found, create a virtual device.
            # see sane-test(5) and /etc/sane.d/test.conf
            pyinsane2.Scanner("test")._open()
            devices = pyinsane2.get_devices()
        self.assertTrue(len(devices) > 0)

    def tearDown(self):
        pyinsane2.exit()


class TestSaneOptions(unittest.TestCase):
    module = None

    def setUp(self):
        pyinsane2.init()
        self.devices = pyinsane2.get_devices()
        self.assertTrue(len(self.devices) > 0)

    def test_get_option(self):
        for dev in self.devices:
            for (k, v) in dev.options.items():
                if v.capabilities.is_active():
                    self.assertNotEqual(v.value, None)
                else:
                    self.assertRaises(
                        pyinsane2.PyinsaneException, lambda: v.value
                    )

    def test_set_option(self):
        for dev in self.devices:
            dev.options['mode'].value = "Gray"
            val = dev.options['mode'].value
            self.assertEqual(val, "Gray")

    def __set_opt(self, dev, opt_name, opt_val):
        dev.options[opt_name].value = opt_val

    def test_set_inexisting_option(self):
        for dev in self.devices:
            self.assertRaises(KeyError, self.__set_opt, dev, 'xyz', "Gray")

    def test_set_invalid_value(self):
        for dev in self.devices:
            self.assertRaises(
                pyinsane2.PyinsaneException,
                self.__set_opt, dev, 'mode', "XYZ"
            )

    def test_set_inactive_option(self):
        for dev in self.devices:
            noncolor = [
                x for x in dev.options["mode"].constraint if x != "Color"
            ]
            if len(noncolor) == 0:
                self.skipTest("scanner does not support required option")
            if "three-pass" not in dev.options:
                self.skipTest("scanner does not support option 'three-pass'")
            dev.options["mode"].value = noncolor[0]
            # three-pass mode is only active in color mode
            self.assertRaises(
                pyinsane2.PyinsaneException,
                self.__set_opt, dev, 'three-pass', 1
            )

    def tearDown(self):
        for dev in self.devices:
            del(dev)
        del(self.devices)
        pyinsane2.exit()


class TestSaneScan(unittest.TestCase):
    module = None

    def setUp(self):
        pyinsane2.init()
        devices = pyinsane2.get_devices()
        self.assertTrue(len(devices) > 0)
        self.dev = devices[0]

    def test_simple_scan_lineart(self):
        try:
            self.dev.options['mode'].value = "Lineart"
        except pyinsane2.PyinsaneException:
            self.dev.options['mode'].value = "Gray"
            self.dev.options['depth'].value = 1
        scan_session = self.dev.scan(multiple=False)
        try:
            assert(scan_session.scan is not None)
            while True:
                scan_session.scan.read()
        except EOFError:
            pass
        img = scan_session.images[0]
        self.assertNotEqual(img, None)

    def test_simple_scan_gray(self):
        try:
            self.dev.options['mode'].value = "Gray"
        except pyinsane2.PyinsaneException:
            self.skipTest("scanner does not support required option")
        scan_session = self.dev.scan(multiple=False)
        try:
            while True:
                scan_session.scan.read()
        except EOFError:
            pass
        img = scan_session.images[0]
        self.assertNotEqual(img, None)

    def test_simple_scan_color(self):
        try:
            self.dev.options['mode'].value = "Color"
        except pyinsane2.PyinsaneException:
            self.skipTest("scanner does not support required option")
        self.dev.options['resolution'].value = 300
        scan_session = self.dev.scan(multiple=False)
        try:
            last_line = 0
            while True:
                scan_session.scan.read()
                nb_lines = scan_session.scan.available_lines[1]
                if nb_lines > last_line + 100:
                    # just making sure it doesn't raise exceptions
                    scan_session.scan.get_image(end_line=nb_lines)
                    last_line = nb_lines
        except EOFError:
            pass
        img = scan_session.images[0]
        self.assertNotEqual(img, None)

    def test_multi_scan_on_flatbed(self):
        try:
            self.dev.options['source'].value = "Flatbed"
            self.dev.options['mode'].value = "Color"
        except pyinsane2.PyinsaneException:
            self.skipTest("scanner does not support required option")
        scan_session = self.dev.scan(multiple=True)
        try:
            while True:
                scan_session.scan.read()
        except EOFError:
            pass
        self.assertEqual(len(scan_session.images), 1)
        self.assertNotEqual(scan_session.images[0], None)

    def test_multi_scan_on_adf(self):
        adf_found = False
        pages = 0
        if "ADF" in self.dev.options['source'].constraint:
            self.dev.options['source'].value = "ADF"
            adf_found = True
        else:
            for srcname in self.dev.options['source'].constraint:
                # sane-test uses 'Automatic Document Feeder' instead of ADF
                # WIA uses 'feeder'
                if "feeder" in srcname.lower():
                    self.dev.options['source'].value = srcname
                    if os.name != "nt":
                        pages = 10  # sane-test scans give us 10 pages
                    adf_found = True
        self.assertTrue(adf_found)
        self.dev.options['mode'].value = "Color"
        if not adf_found:
            self.skipTest("scanner does not support required option")
            return
        scan_session = self.dev.scan(multiple=True)
        try:
            while True:
                try:
                    scan_session.scan.read()
                except EOFError:
                    pass
        except StopIteration:
            pass
        self.assertEqual(len(scan_session.images), pages)

    def test_expected_size(self):
        try:
            self.dev.options['source'].value = "Flatbed"
            self.dev.options['mode'].value = "Color"
        except pyinsane2.PyinsaneException:
            self.skipTest("scanner does not support required option")
        scan_session = self.dev.scan(multiple=False)
        scan_size = scan_session.scan.expected_size
        self.assertTrue(scan_size[0] > 100)
        self.assertTrue(scan_size[1] > 100)
        scan_session.scan.cancel()

    def test_get_progressive_scan(self):
        try:
            self.dev.options['source'].value = "Flatbed"
            self.dev.options['mode'].value = "Color"
        except pyinsane2.PyinsaneException:
            self.skipTest("scanner does not support required option")
        scan_session = self.dev.scan(multiple=False)
        last_line = 0
        expected_size = scan_session.scan.expected_size
        try:
            while True:
                scan_session.scan.read()

                self.assertEqual(scan_session.scan.available_lines[0], 0)
                current_line = scan_session.scan.available_lines[1]
                self.assertTrue(last_line <= current_line)

                if current_line >= last_line + 100:
                    img = scan_session.scan.get_image(0, current_line)
                    self.assertEqual(img.size[0], expected_size[0])
                    self.assertEqual(img.size[1], current_line)
                    last_line = current_line
        except EOFError:
            pass
        self.assertTrue(last_line <= current_line)
        self.assertEqual(current_line, expected_size[1])
        last_line = current_line
        img = scan_session.scan.get_image(0, current_line)
        self.assertEqual(img.size[0], expected_size[0])
        self.assertEqual(img.size[1], current_line)

        img = scan_session.images[0]
        self.assertNotEqual(img, None)

    def tearDown(self):
        del(self.dev)
        pyinsane2.exit()



import os
import unittest

if os.name != "nt":
    from pyinsane2.sane import rawapi


def get_test_devices():
    '''Return SANE devices, perhaps after creating a test device.'''
    devices = rawapi.sane_get_devices()
    if len(devices) == 0:
        # if there are no devices found, create a virtual device.
        # see sane-test(5) and /etc/sane.d/test.conf
        rawapi.sane_close(rawapi.sane_open("test"))
        devices = rawapi.sane_get_devices()
    return devices


class TestSaneInit(unittest.TestCase):
    def setUp(self):
        pass

    @unittest.skipIf(os.name == "nt", "sane only")
    def test_init(self):
        version = rawapi.sane_init()
        self.assertTrue(version.is_current())
        rawapi.sane_exit()

    def tearDown(self):
        pass


class TestSaneGetDevices(unittest.TestCase):
    def setUp(self):
        rawapi.sane_init()

    @unittest.skipIf(os.name == "nt", "sane only")
    def test_get_devices(self):
        devices = get_test_devices()
        self.assertTrue(len(devices) > 0)

    def tearDown(self):
        rawapi.sane_exit()


class TestSaneOpen(unittest.TestCase):
    def setUp(self):
        rawapi.sane_init()
        devices = get_test_devices()
        self.assertTrue(len(devices) > 0)
        self.dev_name = devices[0].name

    @unittest.skipIf(os.name == "nt", "sane only")
    def test_open_invalid(self):
        self.assertRaises(rawapi.SaneException, rawapi.sane_open, "whatever")

    @unittest.skipIf(os.name == "nt", "sane only")
    def test_open_valid(self):
        dev_handle = rawapi.sane_open(self.dev_name)
        rawapi.sane_close(dev_handle)

    def tearDown(self):
        rawapi.sane_exit()


class TestSaneGetOptionDescriptor(unittest.TestCase):
    def setUp(self):
        rawapi.sane_init()
        devices = get_test_devices()
        self.assertTrue(len(devices) > 0)
        dev_name = devices[0].name
        self.dev_handle = rawapi.sane_open(dev_name)

    @unittest.skipIf(os.name == "nt", "sane only")
    def test_get_option_descriptor_0(self):
        opt_desc = rawapi.sane_get_option_descriptor(self.dev_handle, 0)
        # XXX(Jflesch): The name may vary: sometimes it's empty, sometimes it's
        # "option-cnt"
        #self.assertEqual(opt_desc.name, "")
        self.assertEqual(opt_desc.title, b"Number of options")
        self.assertEqual(opt_desc.type, rawapi.SaneValueType.INT)
        self.assertEqual(opt_desc.unit, rawapi.SaneUnit.NONE)
        self.assertEqual(opt_desc.size, 4)
        self.assertEqual(opt_desc.cap, rawapi.SaneCapabilities.SOFT_DETECT)
        self.assertEqual(opt_desc.constraint_type,
                         rawapi.SaneConstraintType.NONE)

    @unittest.skipIf(os.name == "nt", "sane only")
    def test_get_option_descriptor_out_of_bounds(self):
        # XXX(Jflesch): Sane's documentation says get_option_descriptor()
        # should return NULL if the index value is invalid. It seems the actual
        # implementation prefers to segfault.

        #self.assertRaises(rawapi.SaneException,
        #                  rawapi.sane_get_option_descriptor, self.dev_handle,
        #                  999999)
        pass

    def tearDown(self):
        rawapi.sane_close(self.dev_handle)
        rawapi.sane_exit()


class TestSaneControlOption(unittest.TestCase):
    def setUp(self):
        rawapi.sane_init()
        devices = get_test_devices()
        self.assertTrue(len(devices) > 0)
        dev_name = devices[0].name
        self.dev_handle = rawapi.sane_open(dev_name)
        self.nb_options = rawapi.sane_get_option_value(self.dev_handle, 0)

    @unittest.skipIf(os.name == "nt", "sane only")
    def test_get_option_value(self):
        for opt_idx in range(0, self.nb_options):
            desc = rawapi.sane_get_option_descriptor(self.dev_handle, opt_idx)
            if not rawapi.SaneValueType(desc.type).can_getset_opt():
                continue
            if desc.cap|rawapi.SaneCapabilities.INACTIVE == desc.cap:
                continue
            val = rawapi.sane_get_option_value(self.dev_handle, opt_idx)
            self.assertNotEqual(val, None)

    @unittest.skipIf(os.name == "nt", "sane only")
    def test_set_option_value(self):
        for opt_idx in range(0, self.nb_options):
            desc = rawapi.sane_get_option_descriptor(self.dev_handle, opt_idx)
            if (desc.name != "mode"
                or not rawapi.SaneValueType(desc.type).can_getset_opt()):
                continue
            info = rawapi.sane_set_option_value(self.dev_handle, opt_idx, "Gray")
            self.assertFalse(rawapi.SaneInfo.INEXACT in info)
            val = rawapi.sane_get_option_value(self.dev_handle, opt_idx)
            self.assertEqual(val, "Gray")

    @unittest.skipIf(os.name == "nt", "sane only")
    def test_set_option_auto(self):
        # TODO(Jflesch)
        pass

    def tearDown(self):
        rawapi.sane_close(self.dev_handle)
        rawapi.sane_exit()


class TestSaneScan(unittest.TestCase):
    def setUp(self):
        rawapi.sane_init()
        devices = get_test_devices()
        self.assertTrue(len(devices) > 0)
        dev_name = devices[0].name
        self.dev_handle = rawapi.sane_open(dev_name)

    @unittest.skipIf(os.name == "nt", "sane only")
    def test_simple_scan(self):
        # XXX(Jflesch): set_io_mode() always return SANE_STATUS_UNSUPPORTED
        # with my scanner
        #rawapi.sane_set_io_mode(self.dev_handle, non_blocking=False)

        try:
            rawapi.sane_start(self.dev_handle)
        except StopIteration:
            self.skipTest("cannot scan, no document loaded")

        # XXX(Jflesch): get_select_fd() always return SANE_STATUS_UNSUPPORTED
        # with my scanner
        #fd = rawapi.sane_get_select_fd(self.dev_handle)
        #self.assertTrue(fd > 0)

        try:
            while True:
                buf = rawapi.sane_read(self.dev_handle, 128*1024)
                self.assertTrue(len(buf) > 0)
        except EOFError:
            pass
        rawapi.sane_cancel(self.dev_handle)

    @unittest.skipIf(os.name == "nt", "sane only")
    def test_cancelled_scan(self):
        try:
            rawapi.sane_start(self.dev_handle)
        except StopIteration:
            self.skipTest("cannot scan, no document loaded")
        buf = rawapi.sane_read(self.dev_handle, 128*1024)
        self.assertTrue(len(buf) > 0)
        rawapi.sane_cancel(self.dev_handle)

    def tearDown(self):
        rawapi.sane_close(self.dev_handle)
        rawapi.sane_exit()



#!/usr/bin/env python3

import sys

from PIL import Image

import pyinsane2


def main():
    steps = False

    args = sys.argv[1:]
    if len(args) <= 0 or args[0] == "-h" or args[0] == "--help":
        print("Syntax:")
        print("  %s [-s] <output file (JPG)>" % sys.argv[0])
        print("")
        print("Options:")
        print("  -s : Generate intermediate images (may generate a lot of"
              " images !)")
        sys.exit(1)

    for arg in args[:]:
        if arg == "-s":
            steps = True
            args.remove(arg)

    output_file = args[0]
    print("Output file: %s" % output_file)

    print("Looking for scanners ...")
    devices = pyinsane2.get_devices()
    if (len(devices) <= 0):
        print("No scanner detected !")
        sys.exit(1)
    print("Devices detected:")
    print("- " + "\n- ".join([str(d) for d in devices]))

    print("")

    device = devices[0]
    print("Will use: %s" % str(device))

    print("")

    # For the possible resolutions, look at
    #   device.options['resolution'].constraint
    # It will either be:
    # - None: unknown
    # - a tuple: (min resolution, max resolution)
    # - a list: [75, 150, 300, 600, 1200, ...]

    pyinsane2.set_scanner_opt(device, 'source', ['Auto', 'FlatBed'])
    pyinsane2.set_scanner_opt(device, 'resolution', [300])
    try:
        pyinsane2.maximize_scan_area(device)
    except Exception  as exc:
        print("Failed to maximize scan area: {}".format(exc))
    # Beware: Some scanner have "Lineart" or "Gray" as default mode
    pyinsane2.set_scanner_opt(device, 'mode', ['Color'])

    print("")

    print("Scanning ...  ")
    scan_session = device.scan(multiple=False)

    if steps and scan_session.scan.expected_size[1] < 0:
        print("Warning: requested step by step scan images, but"
              " scanner didn't report the expected number of lines"
              " in the final image --> can't do")
        print("Step by step scan images won't be recorded")
        steps = False

    if steps:
        last_line = 0
        expected_size = scan_session.scan.expected_size
        img = Image.new("RGB", expected_size, "#ff00ff")
        sp = output_file.split(".")
        steps_filename = (".".join(sp[:-1]), sp[-1])

    try:
        PROGRESSION_INDICATOR = ['|', '/', '-', '\\']
        i = -1
        while True:
            i += 1
            i %= len(PROGRESSION_INDICATOR)
            sys.stdout.write("\b%s" % PROGRESSION_INDICATOR[i])
            sys.stdout.flush()

            scan_session.scan.read()

            if steps:
                next_line = scan_session.scan.available_lines[1]
                if (next_line > last_line + 100):
                    subimg = scan_session.scan.get_image(last_line, next_line)
                    img.paste(subimg, (0, last_line))
                    img.save("%s-%05d.%s" % (steps_filename[0], last_line,
                                             steps_filename[1]), "JPEG")
                    last_line = next_line
    except EOFError:
        pass

    print("\b ")
    print("Writing output file ...")
    img = scan_session.images[0]
    img.save(output_file, "JPEG")
    print("Done")

if __name__ == "__main__":
    pyinsane2.init()
    try:
        main()
    finally:
        pyinsane2.exit()
